// mp3player.c: MP3 Player Implementation in C for Sony PSP
//
////////////////////////////////////////////////////////////////////////////

#include <pspkernel.h>
#include <pspdebug.h>
#include <pspiofilemgr.h>
#include <pspdisplay.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <pspaudiolib.h>
#include <math.h>
#include "mp3player.h"

#define FALSE 0
#define TRUE !FALSE
#define min(a,b) (((a)<(b))?(a):(b))
#define max(a,b) (((a)>(b))?(a):(b))
#define MadErrorString(x) mad_stream_errorstr(x)
#define INPUT_BUFFER_SIZE	(5*8192)
#define OUTPUT_BUFFER_SIZE	2048	/* Must be an integer multiple of 4. */

/* This table represents the subband-domain filter characteristics. It
* is initialized by the ParseArgs() function and is used as
* coefficients against each subband samples when DoFilter is non-nul.
*/
mad_fixed_t Filter[32];

/* DoFilter is non-nul when the Filter table defines a filter bank to
* be applied to the decoded audio subbands.
*/
int DoFilter = 0;
/* Define printf, just to make typing easier */
#define printf	pspDebugScreenPrintf

#define NUMCHANNELS 2
u8 *ptr;
long size;
long samplesInOutput = 0;

//////////////////////////////////////////////////////////////////////
// Global local variables
//////////////////////////////////////////////////////////////////////

//libmad lowlevel stuff

// The following variables contain the music data, ie they don't change value until you load a new file
struct mad_stream Stream;
struct mad_frame Frame;
struct mad_synth Synth;
mad_timer_t Timer;
signed short OutputBuffer[OUTPUT_BUFFER_SIZE];
unsigned char InputBuffer[INPUT_BUFFER_SIZE + MAD_BUFFER_GUARD],
    *OutputPtr = (unsigned char *) OutputBuffer, *GuardPtr = NULL;
const unsigned char *OutputBufferEnd = (unsigned char *) OutputBuffer + OUTPUT_BUFFER_SIZE * 2;
int Status = 0, i;
unsigned long FrameCount = 0;

// The following variables are maintained and updated by the tracker during playback
static int isPlaying;		// Set to true when a mod is being played

//////////////////////////////////////////////////////////////////////
// These are the public functions
//////////////////////////////////////////////////////////////////////
static int myChannel;
static int eos;

/* Define printf, just to make typing easier */
#define printf	pspDebugScreenPrintf

struct MP3Info MP3_info;
unsigned int MP3_volume_boost = 0;

/****************************************************************************
* Applies a frequency-domain filter to audio data in the subband-domain.	*
****************************************************************************/
static void ApplyFilter(struct mad_frame *Frame)
{
    int Channel, Sample, Samples, SubBand;

    /* There is two application loops, each optimized for the number
     * of audio channels to process. The first alternative is for
     * two-channel frames, the second is for mono-audio.
     */
    Samples = MAD_NSBSAMPLES(&Frame->header);
    if (Frame->header.mode != MAD_MODE_SINGLE_CHANNEL)
	for (Channel = 0; Channel < 2; Channel++)
	    for (Sample = 0; Sample < Samples; Sample++)
		for (SubBand = 0; SubBand < 32; SubBand++)
		    Frame->sbsample[Channel][Sample][SubBand] =
			mad_f_mul(Frame->sbsample[Channel][Sample][SubBand], Filter[SubBand]);
    else
	for (Sample = 0; Sample < Samples; Sample++)
	    for (SubBand = 0; SubBand < 32; SubBand++)
		Frame->sbsample[0][Sample][SubBand] = mad_f_mul(Frame->sbsample[0][Sample][SubBand], Filter[SubBand]);
}

/****************************************************************************
* Converts a sample from libmad's fixed point number format to a signed	*
* short (16 bits).															*
****************************************************************************/
static signed short MadFixedToSshort(mad_fixed_t Fixed)
{
    /* A fixed point number is formed of the following bit pattern:
     *
     * SWWWFFFFFFFFFFFFFFFFFFFFFFFFFFFF
     * MSB                          LSB
     * S ==> Sign (0 is positive, 1 is negative)
     * W ==> Whole part bits
     * F ==> Fractional part bits
     *
     * This pattern contains MAD_F_FRACBITS fractional bits, one
     * should alway use this macro when working on the bits of a fixed
     * point number. It is not guaranteed to be constant over the
     * different platforms supported by libmad.
     *
     * The signed short value is formed, after clipping, by the least
     * significant whole part bit, followed by the 15 most significant
     * fractional part bits. Warning: this is a quick and dirty way to
     * compute the 16-bit number, madplay includes much better
     * algorithms.
     */

    /* Clipping */
    if (Fixed >= MAD_F_ONE)
	return (SHRT_MAX);
    if (Fixed <= -MAD_F_ONE)
	return (-SHRT_MAX);

    /* Conversion. */
    Fixed = Fixed >> (MAD_F_FRACBITS - 15);
    return ((signed short) Fixed);
}

//Volume boost for a single sample:
short volume_boost(short Sample, int boost){
	int intSample = Sample;
	intSample *= (MP3_volume_boost + 1);
	if (intSample > 32767){
		intSample = 32767;
	}else if (intSample < -32768){
		intSample = -32768;
	}
	Sample = intSample;
	return Sample;
}

//MP3 Callback for audio:
static void MP3Callback(void *_buf2, unsigned int numSamples, void *pdata)
{
  short *_buf = (short *)_buf2;
    unsigned long samplesOut = 0;
    //      u8 justStarted = 1;

    if (isPlaying == TRUE) {	//  Playing , so mix up a buffer
		if (samplesInOutput > 0) {
			if (samplesInOutput > numSamples) {
			memcpy((char *) _buf, (char *) OutputBuffer, numSamples * 2 * 2);
			samplesOut = numSamples;
			samplesInOutput -= numSamples;
			} else {
			memcpy((char *) _buf, (char *) OutputBuffer, samplesInOutput * 2 * 2);
			samplesOut = samplesInOutput;
			samplesInOutput = 0;
			}
		}
		while (samplesOut < numSamples) {
			if (Stream.buffer == NULL || Stream.error == MAD_ERROR_BUFLEN) {
				mad_stream_buffer(&Stream, ptr, size);
				Stream.error = 0;
			}

			if (mad_frame_decode(&Frame, &Stream)) {
			if (MAD_RECOVERABLE(Stream.error)) {
				if (Stream.error != MAD_ERROR_LOSTSYNC || Stream.this_frame != GuardPtr) {
					//printf("recoverable frame level error (%s)\n", MadErrorString(&Stream));
					//sceDisplayWaitVblankStart();
				}
				return;
			} else if (Stream.error == MAD_ERROR_BUFLEN) {
				eos = 1;
				return;
			} else {
				Status = 1;
				MP3_Stop();
			}
			}

			if (FrameCount == 0){
				Status = 1;
			}

			FrameCount++;
			mad_timer_add(&Timer, Frame.header.duration);

			if (DoFilter)
				ApplyFilter(&Frame);

			mad_synth_frame(&Synth, &Frame);

			for (i = 0; i < Synth.pcm.length; i++) {
			signed short Sample;
			if (samplesOut < numSamples) {
				/* Left channel */
				Sample = MadFixedToSshort(Synth.pcm.samples[0][i]);
				//Volume Boost:
				if (MP3_volume_boost > 0){
					Sample = volume_boost(Sample, MP3_volume_boost);
				}
				_buf[samplesOut * 2] = Sample;

				/* Right channel. If the decoded stream is monophonic then
				 * the right output channel is the same as the left one.
				 */
				if (MAD_NCHANNELS(&Frame.header) == 2)
					Sample = MadFixedToSshort(Synth.pcm.samples[1][i]);
					//Volume Boost:
					if (MP3_volume_boost > 0){
						Sample = volume_boost(Sample, MP3_volume_boost);
					}
				_buf[samplesOut * 2 + 1] = Sample;
				samplesOut++;
			} else {
				Sample = MadFixedToSshort(Synth.pcm.samples[0][i]);
				if (MP3_volume_boost > 0){
					Sample = volume_boost(Sample, MP3_volume_boost);
				}
				OutputBuffer[samplesInOutput * 2] = Sample;
				if (MAD_NCHANNELS(&Frame.header) == 2)
					Sample = MadFixedToSshort(Synth.pcm.samples[1][i]);
					if (MP3_volume_boost > 0){
						Sample = volume_boost(Sample, MP3_volume_boost);
					}
				OutputBuffer[samplesInOutput * 2 + 1] = Sample;
				samplesInOutput++;
			}

			}
		}
		} else {			//  Not Playing , so clear buffer
		{
			int count;
			for (count = 0; count < numSamples * 2; count++)
			*(_buf + count) = 0;
		}
    }
}

void MP3_Init(int channel)
{
    myChannel = channel;
    isPlaying = FALSE;
	MP3_volume_boost = 0;
    pspAudioSetChannelCallback(myChannel, MP3Callback,0);
    /* First the structures used by libmad must be initialized. */
    mad_stream_init(&Stream);
    mad_frame_init(&Frame);
    mad_synth_init(&Synth);
    mad_timer_reset(&Timer);
}


void MP3_FreeTune()
{
    /* The input file was completely read; the memory allocated by our
     * reading module must be reclaimed.
     */
    if (ptr)
		free(ptr);

    /* Mad is no longer used, the structures that were initialized must
     * now be cleared.
     */
    mad_synth_finish(&Synth);
    mad_frame_finish(&Frame);
    mad_stream_finish(&Stream);

    /* Accounting report if no error occurred. */
    if (!Status) {
		char Buffer[80];

	mad_timer_string(Timer, Buffer, "%lu:%02lu.%03u", MAD_UNITS_MINUTES, MAD_UNITS_MILLISECONDS, 0);
	sceKernelDelayThread(500000);
    }
}


//Recupero le informazioni sul file:
void getInfo(){
	int FrameCount = 0;
	struct mad_stream stream;
	struct mad_header header; 
	mad_stream_init (&stream);
	mad_header_init (&header);

	MP3_info.fileSize = size;
	mad_timer_reset(&MP3_info.length);
	mad_stream_buffer (&stream, ptr, size);
	while (1){
		if (mad_header_decode (&header, &stream) == -1){
			if (MAD_RECOVERABLE(stream.error)){
				continue;				
			}else{
				break;
			}
		}
		//Informazioni solo dal primo frame:
	    if (FrameCount == 0){
			switch (header.layer) {
			case MAD_LAYER_I:
				strcpy(MP3_info.layer,"I");
				break;
			case MAD_LAYER_II:
				strcpy(MP3_info.layer,"II");
				break;
			case MAD_LAYER_III:
				strcpy(MP3_info.layer,"III");
				break;
			default:
				strcpy(MP3_info.layer,"unknown");
				break;
			}

			MP3_info.kbit = header.bitrate / 1000;
			MP3_info.hz = header.samplerate;
			switch (header.mode) {
			case MAD_MODE_SINGLE_CHANNEL:
				strcpy(MP3_info.mode, "single channel");
				break;
			case MAD_MODE_DUAL_CHANNEL:
				strcpy(MP3_info.mode, "dual channel");
				break;
			case MAD_MODE_JOINT_STEREO:
				strcpy(MP3_info.mode, "joint (MS/intensity) stereo");
				break;
			case MAD_MODE_STEREO:
				strcpy(MP3_info.mode, "normal LR stereo");
				break;
			default:
				strcpy(MP3_info.mode, "unknown");
				break;
			}

			switch (header.emphasis) {
			case MAD_EMPHASIS_NONE:
				strcpy(MP3_info.emphasis,"no");
				break;
			case MAD_EMPHASIS_50_15_US:
				strcpy(MP3_info.emphasis,"50/15 us");
				break;
			case MAD_EMPHASIS_CCITT_J_17:
				strcpy(MP3_info.emphasis,"CCITT J.17");
				break;
			case MAD_EMPHASIS_RESERVED:
				strcpy(MP3_info.emphasis,"reserved(!)");
				break;
			default:
				strcpy(MP3_info.emphasis,"unknown");
				break;
			}			
		}
		//Conteggio frame e durata totale:
		FrameCount++;
		mad_timer_add (&MP3_info.length, header.duration);
	}
	mad_header_finish (&header);
	mad_stream_finish (&stream);

	MP3_info.frames = FrameCount;
	//Formatto in stringa la durata totale:
	mad_timer_string(MP3_info.length, MP3_info.strLength, "%02lu:%02u:%02u", MAD_UNITS_HOURS, MAD_UNITS_MILLISECONDS, 0);
}

//MP3_End
void MP3_End()
{
    MP3_Stop();
    pspAudioSetChannelCallback(myChannel, 0,0);
    MP3_FreeTune();
}

//////////////////////////////////////////////////////////////////////
// Functions - Local and not public
//////////////////////////////////////////////////////////////////////

//  This is the initialiser and module loader
//  This is a general call, which loads the module from the 
//  given address into the modplayer
//
//  It basically loads into an internal format, so once this function
//  has returned the buffer at 'data' will not be needed again.
int MP3_Load(char *filename)
{
    int fd;
    eos = 0;
    if ((fd = sceIoOpen(filename, PSP_O_RDONLY, 0777)) > 0) {
		//  opened file, so get size now
		size = sceIoLseek(fd, 0, PSP_SEEK_END);
		sceIoLseek(fd, 0, PSP_SEEK_SET);
		ptr = (unsigned char *) malloc(size + 8);
		memset(ptr, 0, size + 8);
		if (ptr) {		// Read file in
			sceIoRead(fd, ptr, size);
		} else {
			//printf("Error allocing\n");
			sceIoClose(fd);
			return 0;
		}
		// Close file
		sceIoClose(fd);
    } else {
		return 0;
    }
    isPlaying = FALSE;
	FrameCount = 0;

	getInfo();
    return 1;
}

// This function initialises for playing, and starts
int MP3_Play()
{
	//Azzero il timer:
    if (eos == 1){
		mad_timer_reset(&Timer);
	}

	// See if I'm already playing
    if (isPlaying)
		return FALSE;

    isPlaying = TRUE;

	return TRUE;
}

void MP3_Pause()
{
    isPlaying = !isPlaying;
}

int MP3_Stop()
{
    //stop playing
    isPlaying = FALSE;

    //clear buffer
    memset(OutputBuffer, 0, OUTPUT_BUFFER_SIZE);
    OutputPtr = (unsigned char *) OutputBuffer;

    return TRUE;
}

void MP3_GetTimeString(char *dest)
{
    mad_timer_string(Timer, dest, "%02lu:%02u:%02u", MAD_UNITS_HOURS, MAD_UNITS_MILLISECONDS, 0);
}

int MP3_GetPercentace()
{
	//Calcolo posizione in %:
	float perc;
	 
	if (mad_timer_count(MP3_info.length, MAD_UNITS_SECONDS) > 0){
		perc = ((float)mad_timer_count(Timer, MAD_UNITS_SECONDS) / (float)mad_timer_count(MP3_info.length, MAD_UNITS_SECONDS)) * 100.0;
	}else{
		perc = 0;
	}
    return((int)perc);
}

int MP3_EndOfStream()
{
    if (eos == 1)
	return 1;
    return 0;
}

//Get info on file:
struct MP3Info MP3_GetInfo()
{
	return MP3_info;
}

int MP3_GetStatus(){
	return (0);
}

//Set volume boost:
void MP3_setVolumeBoost(int boost){
	MP3_volume_boost = boost;
}

//Get actual volume boost:
int MP3_getVolumeBoost(){
	return(MP3_volume_boost);
}

//Set Filter:
int MP3_setFilter(double tFilter[32]){
	//Converto i db:
	double AmpFactor;
	int i;

	for (i = 0; i < 32; i++){
		AmpFactor=pow(10.,tFilter[i]/20);
		if(AmpFactor>mad_f_todouble(MAD_F_MAX))
		{
			DoFilter = 0;
			return(1);
		}

		/* Eventually the amplification factor is converted
		 * from double to fixed point with mad_f_tofixed().
		 */
		Filter[i]=mad_f_tofixed(AmpFactor);
	}
	return(0);
}

//Check if filter is enabled:
int MP3_isFilterEnabled(){
	return DoFilter;
}

//Enable filter:
void MP3_enableFilter(){
	DoFilter = 1;
}

//Disable filter:
void MP3_disableFilter(){
	DoFilter = 0;
}
